#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest
from time import sleep

from aura_engine.base.config import AuraConfig
from aura_engine.base.logger import AuraLogger
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.core.channels import PlayoutStatusResponse
from aura_engine.engine import Engine, Player
from aura_engine.events import EngineEventDispatcher
from aura_engine.scheduling.domain import Playlist, PlaylistItem, Show, Timeslot
from aura_engine.scheduling.scheduler import AuraScheduler, PlayCommand, TimeslotCommand

from .core_client_mock import PlayoutClientMock

#
# Mocks
#


class MockedEngine:
    """
    Mocked version of engine.
    """

    scheduler: AuraScheduler
    player: Player

    def __init__(self):
        """
        Init.
        """
        self.scheduler = None
        self.player = Player()


class TestSchedulingSchedulerCoreMock(unittest.TestCase):
    """
    Testing the scheduling utils.
    """

    engine: Engine
    client: PlayoutClientMock
    scheduler: AuraScheduler

    # Setup and teardown

    def setUp(self):
        config = AuraConfig.instance.config
        AuraLogger(config)

        engine = Engine.get_instance()
        dispatcher = EngineEventDispatcher(engine=engine)
        client = PlayoutClientMock(event_dispatcher=dispatcher)
        scheduler = AuraScheduler(engine=engine)

        engine.event_dispatcher = dispatcher
        engine.player = Player(client, dispatcher)

        self.engine = engine
        self.client = client
        self.scheduler = scheduler

    def tearDown(self):
        pass

    #
    # Utilities
    #

    def build_timetable(
        self, current_start: int, current_end: int, next_duration: int = 5
    ) -> list[Timeslot]:
        # Delete the timetable file to isolate test case
        f = self.scheduler.timetable.timetable_file
        self.scheduler.timetable.timetable_file = f.replace(
            "timetable.json", "timetable-test.json"
        )
        #

        # Build some timetable
        now = SU.timestamp()
        ts1 = Timeslot(
            id=1,
            repetition_id=None,
            start=now,
            end=now + current_start,
            show=Show(id="1", name="Warmup Show"),
            episode=None,
        )
        ts2 = Timeslot(
            id=2,
            repetition_id=None,
            start=now + current_start,
            end=now + current_end,
            show=Show(id="2", name="Main Show"),
            episode=None,
        )
        ts3 = Timeslot(
            id=3,
            repetition_id=None,
            start=now + current_end,
            end=now + current_end + next_duration,
            show=Show(id="3", name="After Show"),
            episode=None,
        )

        # Playlist for timeslot 2
        item_duration = float(current_end - current_start) / 2.0
        i2a = PlaylistItem("http://stream.local/", item_duration, 100, None)
        i2b = PlaylistItem("http://stream2.local/", item_duration, 100, None)
        pl2 = Playlist("2222", "Streams,", False)
        pl2.add(i2a)
        pl2.add(i2b)

        # Playlist for timeslot 3
        it3 = PlaylistItem("file://2", next_duration, 100, None)
        pl3 = Playlist("3333", "Files", False)
        pl3.add(it3)

        ts2.set_playlists(pl2, None, None)
        ts3.set_playlists(pl3, None, None)
        return [ts2, ts3]

    def run_scheduling_test(self, current_start: int, current_end: int, next_duration: int):

        timeslots = self.build_timetable(current_start, current_end, next_duration)

        next_timeslot: Timeslot
        for next_timeslot in timeslots:
            # Create command timer to indicate the start of the timeslot
            AuraScheduler.TimeslotCommandClass(self.engine, next_timeslot)
            playlist = next_timeslot.get_current_playlist()
            if playlist:
                self.scheduler.queue_playlist_items(next_timeslot, playlist.items)

        # Finish test when scheduler is done
        sleep(current_start + current_end + next_duration)

    #
    # Test Cases
    #

    def test_queue_playlist_items_immediately(self):
        """
        Tests a scenario where the input stream source can't be reached.
        Expected behavior: try to connect until duration is over; then invalidate item and stop related logging.
        (Background thread should terminate when last timeslot has passed.)
        """
        print(self._testMethodName)

        # Configure client and scheduler
        self.client.stream_status = PlayoutStatusResponse.STREAM_STATUS_POLLING
        self.scheduler.config.scheduler.preload_offset = 10

        # Set timetable
        current_start = 5
        current_end = 65
        next_duration = 5

        # Run test
        self.run_scheduling_test(current_start, current_end, next_duration)

    def test_queue_playlist_items_timer(self):
        """
        Like testcase above but spawns a timer which should finish when next item starts.
        """
        print(self._testMethodName)

        # Configure client and scheduler
        self.client.stream_status = PlayoutStatusResponse.STREAM_STATUS_POLLING
        self.scheduler.config.scheduler.preload_offset = 5

        # Set timetable
        current_start = 10
        current_end = 60
        next_duration = 5

        # Run test
        self.run_scheduling_test(current_start, current_end, next_duration)


if __name__ == "__main__":
    unittest.main()
